import 'package:fsm2/src/transitions/fork_transition.dart';
import 'package:fsm2/src/transitions/on_transition.dart';

import 'fork_builder.dart';
import 'graph_builder.dart';
import '../state_machine.dart';

import '../exceptions.dart';
import '../definitions/state_definition.dart';
import '../transitions/join_transition.dart';
import '../types.dart';

/// State builder.
///
/// Instance of this class is passed to [GraphBuilder.state] method.
class StateBuilder<S extends State> {
  final StateDefinition<S> _stateDefinition;

  /// The initial state for the substate
  /// If there are no child states then this is just 'this'.
  Type _initialState;

  StateBuilder(Type stateType, StateDefinition parent, this._stateDefinition) {
    _stateDefinition.setParent(parent);
  }

  /// Statically declares a transition that will occur when Event of type [E]
  /// is sent to machine via [StateMachine.applyEvent] method.
  ///
  ///
  /// ```dart
  /// ..state<MobileNoAcquired>((builder) => builder
  ///   ..on<OnUserFound, LoggedIn>())
  /// ```
  ///
  /// The [condition] argument implements the UML concept a 'guard condition' and
  /// allows you to register multiple transitions for a single Event.
  /// Guard conditions allow you to implement a UML 'Choice psuedostate'.
  /// When the Event is fired each transition will be evaluated in the order
  /// they are added to the State.
  /// The first transition whose guard [condition] method returns true will be triggered, any later
  /// conditions will not be evaluated.
  ///
  /// ```dart
  /// ..state<MobileNoAcquired>((builder) => builder
  ///   ..on<OnUserFound, LoggedIn>(condition: (state, event) => event.subscribed == true))
  ///   ..on<OnUserFound, AskForSubscription>(condition: (state, event) => event.subscribed == false))
  /// ```a
  ///
  /// There MAY be only one transition with a null [condition] and it MUST be the last
  /// transition added to the [S]. A transition with a null [condition] is considered the
  /// 'else' condition in that it fires if none of the transitions with a [condition] evaluate to true.
  ///
  /// An [NullChoiceMustBeLastException] will be thrown if you try to register two
  /// transitions for a given Event type with a null [condition] or you try to add a
  /// transition with a non-null [condition] after adding a transition with a null [condition].
  void on<E extends Event, TOSTATE extends State>(
      {GuardCondition<E> condition, SideEffect sideEffect}) {
    var onTransition = OnTransitionDefinition<S, E, TOSTATE>(
        _stateDefinition, condition, TOSTATE, sideEffect);

    _stateDefinition.addTransition<E>(onTransition);
  }

  /// Adds a nested State definition as per the UML2
  /// specification for `hierarchically nested states`.
  void state<C extends State>(BuildState<C> buildState,
      {bool pageBreak = false}) {
    _stateDefinition.addNestedState(buildState, pageBreak: pageBreak);
  }

  /// Adds a [coregion] State definition as per the UML2
  /// specification for `orthogonal regions`.
  /// If [pageBreak] is set to true (default is false) then
  /// This coregion and its nested children will be exported to a
  /// separate page.
  void coregion<CO extends State>(BuildCoRegion<CO> buildState,
      {bool pageBreak = false}) {
    _stateDefinition.addCoRegion(buildState, pageBreak: pageBreak);
  }

  /// Used to enter a co-region by targeting the set of states within the
  /// coregion to transition to.
  void onFork<E extends Event>(BuildFork<E> buildFork,
      {Function(State, E) condition}) {
    final builder = ForkBuilder<E>();
    buildFork(builder);
    final definition = builder.build();

    var choice = ForkTransitionDefinition<S, E>(_stateDefinition, definition);

    _stateDefinition.addTransition(choice);
  }

  /// Adds an event to the set of events that must be triggered to leave the parent [coregion].
  /// Every onJoin in a coregion must target the same external state.
  void onJoin<E extends Event, TOSTATE extends State>(
      {GuardCondition<E> condition, SideEffect sideEffect}) {
    var onTransition = JoinTransitionDefinition<S, E, TOSTATE>(
        _stateDefinition, condition, sideEffect);

    _stateDefinition.addTransition<E>(onTransition);
  }

  /// Sets callback that will be called right after machine enters this State.
  void onEnter(OnEnter onEnter, {String label}) {
    _stateDefinition.onEnter = onEnter;
    _stateDefinition.nEnterLabel = label;
  }

  /// Sets callback that will be called right before machine exits this State.
  void onExit(OnExit onExit, {String label}) {
    _stateDefinition.onExit = onExit;
    _stateDefinition.onExitLabel = label;
  }

  StateDefinition build() {
    if (_stateDefinition.isLeaf) {
      _initialState = _stateDefinition.stateType;
      return _stateDefinition;
    } else {
      /// If no initial state then the first state is the initial state.
      if (_initialState == null &&
          _stateDefinition.childStateDefinitions.isNotEmpty) {
        _initialState = _stateDefinition.childStateDefinitions[0].stateType;
      }

      assert(_initialState != null);
      var sd = _stateDefinition.findStateDefintion(_initialState,
          includeChildren: false);
      if (sd == null) {
        throw InvalidInitialStateException(
            'The initialState $_initialState MUST be a child state of ${_stateDefinition.stateType}.');
      }

      _stateDefinition.initialState = _initialState;

      return _stateDefinition;
    }
  }

  void initialState<I extends State>() {
    _initialState = I;
  }
}
